@@ -11,8 +11,8 @@ import {
   HTTPMode,
   LENIENT_FLAGS,
   MAJOR, METHOD_MAP, METHODS, METHODS_HTTP, METHODS_ICE, METHODS_RTSP,
-  MINOR, NUM_MAP, SPECIAL_HEADERS, STRICT_TOKEN,
-  TOKEN, TYPE,
+  MINOR, NUM_MAP, QUOTED_STRING, SPECIAL_HEADERS,
+  STRICT_TOKEN, TOKEN, TYPE,
 } from './constants';
 import { URL } from './url';
 
@@ -20,7 +20,9 @@ type MaybeNode = string | Match | Node;
 
 const NODES: ReadonlyArray<string> = [
   'start',
+  'after_start',
   'start_req',
+  'after_start_req',
   'start_res',
   'start_req_or_res',
 
@@ -30,6 +32,7 @@ const NODES: ReadonlyArray<string> = [
   'res_http_dot',
   'res_http_minor',
   'res_http_end',
+  'res_after_version',
   'res_status_code',
   'res_status_code_otherwise',
   'res_status_start',
@@ -39,6 +42,7 @@ const NODES: ReadonlyArray<string> = [
   'req_first_space_before_url',
   'req_spaces_before_url',
   'req_http_start',
+  'req_http_version',
   'req_http_major',
   'req_http_dot',
   'req_http_minor',
@@ -84,7 +88,11 @@ const NODES: ReadonlyArray<string> = [
   'chunk_size_otherwise',
   'chunk_size_almost_done',
   'chunk_size_almost_done_lf',
-  'chunk_parameters',
+  'chunk_extensions',
+  'chunk_extension_name',
+  'chunk_extension_value',
+  'chunk_extension_quoted_value',
+  'chunk_extension_quoted_value_done',
   'chunk_data',
   'chunk_data_almost_done',
   'chunk_data_almost_done_skip',
@@ -101,25 +109,34 @@ const NODES: ReadonlyArray<string> = [
 ];
 
 interface ISpanMap {
-  readonly body: source.Span;
+  readonly status: source.Span;
+  readonly method: source.Span;
+  readonly version: source.Span;
   readonly headerField: source.Span;
   readonly headerValue: source.Span;
-  readonly status: source.Span;
+  readonly chunkExtensionName: source.Span;
+  readonly chunkExtensionValue: source.Span;
+  readonly body: source.Span;
 }
 
 interface ICallbackMap {
-  readonly afterHeadersComplete: source.code.Code;
-  readonly afterMessageComplete: source.code.Code;
-  readonly beforeHeadersComplete: source.code.Code;
-  readonly onChunkComplete: source.code.Code;
-  readonly onChunkHeader: source.code.Code;
-  readonly onHeadersComplete: source.code.Code;
   readonly onMessageBegin: source.code.Code;
-  readonly onMessageComplete: source.code.Code;
   readonly onUrlComplete: source.code.Code;
+  readonly onMethodComplete: source.code.Code;
+  readonly onVersionComplete: source.code.Code;
   readonly onStatusComplete: source.code.Code;
+  readonly beforeHeadersComplete: source.code.Code;
   readonly onHeaderFieldComplete: source.code.Code;
   readonly onHeaderValueComplete: source.code.Code;
+  readonly onHeadersComplete: source.code.Code;
+  readonly afterHeadersComplete: source.code.Code;
+  readonly onChunkHeader: source.code.Code;
+  readonly onChunkExtensionName: source.code.Code;
+  readonly onChunkExtensionValue: source.code.Code;
+  readonly onChunkComplete: source.code.Code;
+  readonly onMessageComplete: source.code.Code;
+  readonly afterMessageComplete: source.code.Code;
+  readonly onReset: source.code.Code;
 }
 
 interface IMulTargets {
@@ -158,9 +175,13 @@ export class HTTP {
 
     this.span = {
       body: p.span(p.code.span('llhttp__on_body')),
+      chunkExtensionName: p.span(p.code.span('llhttp__on_chunk_extension_name')),
+      chunkExtensionValue: p.span(p.code.span('llhttp__on_chunk_extension_value')),
       headerField: p.span(p.code.span('llhttp__on_header_field')),
       headerValue: p.span(p.code.span('llhttp__on_header_value')),
+      method: p.span(p.code.span('llhttp__on_method')),
       status: p.span(p.code.span('llhttp__on_status')),
+      version: p.span(p.code.span('llhttp__on_version')),
     };
 
     /* tslint:disable:object-literal-sort-keys */
@@ -168,13 +189,18 @@ export class HTTP {
       // User callbacks
       onUrlComplete: p.code.match('llhttp__on_url_complete'),
       onStatusComplete: p.code.match('llhttp__on_status_complete'),
+      onMethodComplete: p.code.match('llhttp__on_method_complete'),
+      onVersionComplete: p.code.match('llhttp__on_version_complete'),
       onHeaderFieldComplete: p.code.match('llhttp__on_header_field_complete'),
       onHeaderValueComplete: p.code.match('llhttp__on_header_value_complete'),
       onHeadersComplete: p.code.match('llhttp__on_headers_complete'),
       onMessageBegin: p.code.match('llhttp__on_message_begin'),
       onMessageComplete: p.code.match('llhttp__on_message_complete'),
       onChunkHeader: p.code.match('llhttp__on_chunk_header'),
+      onChunkExtensionName: p.code.match('llhttp__on_chunk_extension_name_complete'),
+      onChunkExtensionValue: p.code.match('llhttp__on_chunk_extension_value_complete'),
       onChunkComplete: p.code.match('llhttp__on_chunk_complete'),
+      onReset: p.code.match('llhttp__on_reset'),
 
       // Internal callbacks `src/http.c`
       beforeHeadersComplete:
@@ -203,6 +229,7 @@ export class HTTP {
     p.property('i8', 'finish');
     p.property('i16', 'flags');
     p.property('i16', 'status_code');
+    p.property('i8', 'initial_message_completed');
 
     // Verify defaults
     assert.strictEqual(FINISH.SAFE, 0);
@@ -233,58 +260,80 @@ export class HTTP {
 
     n('start')
       .match([ '\r', '\n' ], n('start'))
-      .otherwise(this.update('finish', FINISH.UNSAFE,
-        this.invokePausable('on_message_begin',
-          ERROR.CB_MESSAGE_BEGIN, switchType)));
+      .otherwise(
+        this.load('initial_message_completed', {
+          1: this.invokePausable('on_reset', ERROR.CB_RESET, n('after_start')),
+        }, n('after_start')),
+      );
+
+    n('after_start').otherwise(
+      this.update(
+        'finish',
+        FINISH.UNSAFE,
+        this.invokePausable('on_message_begin', ERROR.CB_MESSAGE_BEGIN, switchType),
+      ),
+    );
 
     n('start_req_or_res')
-      .peek('H', n('req_or_res_method'))
+      .peek('H', this.span.method.start(n('req_or_res_method')))
       .otherwise(this.update('type', TYPE.REQUEST, 'start_req'));
 
     n('req_or_res_method')
       .select(H_METHOD_MAP, this.store('method',
-        this.update('type', TYPE.REQUEST, 'req_first_space_before_url')))
-      .match('HTTP/', this.update('type', TYPE.RESPONSE, 'res_http_major'))
+        this.update('type', TYPE.REQUEST, this.span.method.end(
+          this.invokePausable('on_method_complete', ERROR.CB_METHOD_COMPLETE, n('req_first_space_before_url')),
+        )),
+      ))
+      .match('HTTP/', this.span.method.end(this.update('type', TYPE.RESPONSE,
+        this.span.version.start(n('res_http_major')))))
       .otherwise(p.error(ERROR.INVALID_CONSTANT, 'Invalid word encountered'));
 
     const checkVersion = (destination: string): Node => {
+      const node = n(destination);
+      const errorNode = this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Invalid HTTP version'));
+
       return this.testLenientFlags(LENIENT_FLAGS.VERSION,
         {
-          1: n(destination),
+          1: node,
         },
         this.load('http_major', {
           0: this.load('http_minor', {
-            9: n(destination),
-          }, p.error(ERROR.INVALID_VERSION, 'Invalid HTTP version')),
+            9: node,
+          }, errorNode),
           1: this.load('http_minor', {
-            0: n(destination),
-            1: n(destination),
-          }, p.error(ERROR.INVALID_VERSION, 'Invalid HTTP version')),
+            0: node,
+            1: node,
+          }, errorNode),
           2: this.load('http_minor', {
-            0: n(destination),
-          }, p.error(ERROR.INVALID_VERSION, 'Invalid HTTP version')),
-        }, p.error(ERROR.INVALID_VERSION, 'Invalid HTTP version')),
+            0: node,
+          }, errorNode),
+        }, errorNode),
       );
     };
 
     // Response
     n('start_res')
-      .match('HTTP/', n('res_http_major'))
+      .match('HTTP/', span.version.start(n('res_http_major')))
       .otherwise(p.error(ERROR.INVALID_CONSTANT, 'Expected HTTP/'));
 
     n('res_http_major')
       .select(MAJOR, this.store('http_major', 'res_http_dot'))
-      .otherwise(p.error(ERROR.INVALID_VERSION, 'Invalid major version'));
+      .otherwise(this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Invalid major version')));
 
     n('res_http_dot')
       .match('.', n('res_http_minor'))
-      .otherwise(p.error(ERROR.INVALID_VERSION, 'Expected dot'));
+      .otherwise(this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Expected dot')));
 
     n('res_http_minor')
       .select(MINOR, this.store('http_minor', checkVersion('res_http_end')))
-      .otherwise(p.error(ERROR.INVALID_VERSION, 'Invalid minor version'));
+      .otherwise(this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Invalid minor version')));
 
     n('res_http_end')
+      .otherwise(this.span.version.end().otherwise(
+        this.invokePausable('on_version_complete', ERROR.CB_VERSION_COMPLETE, 'res_after_version'),
+      ));
+
+    n('res_after_version')
       .match(' ', this.update('status_code', 0, 'res_status_code'))
       .otherwise(p.error(ERROR.INVALID_VERSION,
           'Expected space after version'));
@@ -301,8 +350,9 @@ export class HTTP {
       .peek([ '\r', '\n' ], n('res_status_start'))
       .otherwise(p.error(ERROR.INVALID_STATUS, 'Invalid response status'));
 
-    const onStatusComplete = p.invoke(this.callback.onStatusComplete);
-    onStatusComplete.otherwise(n('headers_start'));
+    const onStatusComplete = this.invokePausable(
+      'on_status_complete', ERROR.CB_STATUS_COMPLETE, n('headers_start'),
+    );
 
     n('res_status_start')
       .match('\r', n('res_line_almost_done'))
@@ -324,10 +374,12 @@ export class HTTP {
     }
 
     // Request
+    n('start_req').otherwise(this.span.method.start(n('after_start_req')));
 
-    n('start_req')
-      .select(METHOD_MAP,
-        this.store('method', 'req_first_space_before_url'))
+    n('after_start_req')
+      .select(METHOD_MAP, this.store('method', this.span.method.end(
+        this.invokePausable('on_method_complete', ERROR.CB_METHOD_COMPLETE, n('req_first_space_before_url'),
+      ))))
       .otherwise(p.error(ERROR.INVALID_METHOD, 'Invalid method encountered'));
 
     n('req_first_space_before_url')
@@ -359,7 +411,7 @@ export class HTTP {
       );
 
     const checkMethod = (methods: METHODS[], error: string): Node => {
-      const success = n('req_http_major');
+      const success = n('req_http_version');
       const failure = p.error(ERROR.INVALID_CONSTANT, error);
 
       const map: { [key: number]: Node } = {};
@@ -380,21 +432,30 @@ export class HTTP {
       .match(' ', n('req_http_start'))
       .otherwise(p.error(ERROR.INVALID_CONSTANT, 'Expected HTTP/'));
 
+    n('req_http_version').otherwise(span.version.start(n('req_http_major')));
+
     n('req_http_major')
       .select(MAJOR, this.store('http_major', 'req_http_dot'))
-      .otherwise(p.error(ERROR.INVALID_VERSION, 'Invalid major version'));
+      .otherwise(this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Invalid major version')));
 
     n('req_http_dot')
       .match('.', n('req_http_minor'))
-      .otherwise(p.error(ERROR.INVALID_VERSION, 'Expected dot'));
+      .otherwise(this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Expected dot')));
 
     n('req_http_minor')
       .select(MINOR, this.store('http_minor', checkVersion('req_http_end')))
-      .otherwise(p.error(ERROR.INVALID_VERSION, 'Invalid minor version'));
+      .otherwise(this.span.version.end(p.error(ERROR.INVALID_VERSION, 'Invalid minor version')));
 
-    n('req_http_end').otherwise(this.load('method', {
-      [METHODS.PRI]: n('req_pri_upgrade'),
-    }, n('req_http_complete')));
+    n('req_http_end').otherwise(
+      span.version.end().otherwise(
+        this.invokePausable(
+          'on_version_complete',
+          ERROR.CB_VERSION_COMPLETE,
+          this.load('method', {
+            [METHODS.PRI]: n('req_pri_upgrade'),
+          }, n('req_http_complete')),
+        )),
+    );
 
     n('req_http_complete')
       .match([ '\r\n', '\n' ], n('headers_start'))
@@ -490,7 +551,9 @@ export class HTTP {
     n('header_value_discard_ws')
       .match([ ' ', '\t' ], n('header_value_discard_ws'))
       .match('\r', n('header_value_discard_ws_almost_done'))
-      .match('\n', n('header_value_discard_lws'))
+      .match('\n', this.testLenientFlags(LENIENT_FLAGS.HEADERS, {
+        1: n('header_value_discard_lws'),
+      }, p.error(ERROR.INVALID_HEADER_TOKEN, 'Invalid header value char')))
       .otherwise(span.headerValue.start(n('header_value_start')));
 
     if (this.mode === 'strict') {
@@ -653,7 +716,7 @@ export class HTTP {
 
     const checkLenient = this.testLenientFlags(LENIENT_FLAGS.HEADERS, {
       1: n('header_value_lenient'),
-    }, n('header_value_lenient_failed'));
+    }, span.headerValue.end(p.error(ERROR.INVALID_HEADER_TOKEN, 'Invalid header value char')));
 
     n('header_value_otherwise')
       .peek('\r', span.headerValue.end().skipTo(n('header_value_almost_done')))
@@ -676,7 +739,11 @@ export class HTTP {
         'Missing expected LF after header value'));
 
     n('header_value_lws')
-      .peek([ ' ', '\t' ], span.headerValue.start(n('header_value_start')))
+      .peek([ ' ', '\t' ],
+        this.load('header_state', {
+          [HEADER_STATE.TRANSFER_ENCODING_CHUNKED]:
+            this.resetHeaderState(span.headerValue.start(n('header_value_start'))),
+        }, span.headerValue.start(n('header_value_start'))))
       .otherwise(this.setHeaderFlags(onHeaderValueComplete));
 
     const checkTrailing = this.testFlags(FLAGS.TRAILING, {
@@ -806,15 +873,69 @@ export class HTTP {
 
     n('chunk_size_otherwise')
       .match('\r', n('chunk_size_almost_done'))
-      .match('; ', n('chunk_parameters'))
+      .match(';', n('chunk_extensions'))
       .otherwise(p.error(ERROR.INVALID_CHUNK_SIZE,
         'Invalid character in chunk size'));
 
-    n('chunk_parameters')
+    const onChunkExtensionNameCompleted = (destination: Node) => {
+      return this.invokePausable(
+       'on_chunk_extension_name', ERROR.CB_CHUNK_EXTENSION_NAME_COMPLETE, destination);
+    };
+
+    const onChunkExtensionValueCompleted = (destination: Node) => {
+      return this.invokePausable(
+       'on_chunk_extension_value', ERROR.CB_CHUNK_EXTENSION_VALUE_COMPLETE, destination);
+    };
+
+    n('chunk_extensions')
+      .match(' ', p.error(ERROR.STRICT, 'Invalid character in chunk extensions'))
+      .match('\r', p.error(ERROR.STRICT, 'Invalid character in chunk extensions'))
+      .otherwise(this.span.chunkExtensionName.start(n('chunk_extension_name')));
+
+    n('chunk_extension_name')
+      .match(STRICT_TOKEN, n('chunk_extension_name'))
+      .peek('=', this.span.chunkExtensionName.end().skipTo(
+        this.span.chunkExtensionValue.start(
+          onChunkExtensionNameCompleted(n('chunk_extension_value')),
+        ),
+      ))
+      .peek(';', this.span.chunkExtensionName.end().skipTo(
+        onChunkExtensionNameCompleted(n('chunk_extensions')),
+      ))
+      .peek('\r', this.span.chunkExtensionName.end().skipTo(
+        onChunkExtensionNameCompleted(n('chunk_size_almost_done')),
+      ))
+      .otherwise(this.span.chunkExtensionName.end().skipTo(
+        p.error(ERROR.STRICT, 'Invalid character in chunk extensions name'),
+      ));
+
+    n('chunk_extension_value')
+      .match('"', n('chunk_extension_quoted_value'))
+      .match(STRICT_TOKEN, n('chunk_extension_value'))
+      .peek(';', this.span.chunkExtensionValue.end().skipTo(
+        onChunkExtensionValueCompleted(n('chunk_size_otherwise')),
+      ))
+      .peek('\r', this.span.chunkExtensionValue.end().skipTo(
+        onChunkExtensionValueCompleted(n('chunk_size_almost_done')),
+      ))
+      .otherwise(this.span.chunkExtensionValue.end().skipTo(
+        p.error(ERROR.STRICT, 'Invalid character in chunk extensions value'),
+      ));
+
+    n('chunk_extension_quoted_value')
+      .match(QUOTED_STRING, n('chunk_extension_quoted_value'))
+      .match('"', this.span.chunkExtensionValue.end(
+        onChunkExtensionValueCompleted(n('chunk_extension_quoted_value_done')),
+      ))
+      .otherwise(this.span.chunkExtensionValue.end().skipTo(
+        p.error(ERROR.STRICT, 'Invalid character in chunk extensions quoted value'),
+      ));
+
+    n('chunk_extension_quoted_value_done')
+      .match(';', n('chunk_extensions'))
       .match('\r', n('chunk_size_almost_done'))
-      .match(HEADER_CHARS, n('chunk_parameters'))
       .otherwise(p.error(ERROR.STRICT,
-        'Invalid character in chunk parameters'));
+        'Invalid character in chunk extensions quote value'));
 
     if (this.mode === 'strict') {
       n('chunk_size_almost_done')
@@ -874,7 +995,7 @@ export class HTTP {
     // Check if we'd like to keep-alive
     n('cleanup')
       .otherwise(p.invoke(callback.afterMessageComplete, {
-        1: n('restart'),
+        1: this.update('content_length', 0, n('restart')),
       }, this.update('finish', FINISH.SAFE, lenientClose)));
 
     if (this.mode === 'strict') {
@@ -889,7 +1010,9 @@ export class HTTP {
     }
 
     n('restart')
-      .otherwise(this.update('finish', FINISH.SAFE, n('start')));
+      .otherwise(
+        this.update('initial_message_completed', 1, this.update('finish', FINISH.SAFE, n('start')),
+      ));
   }
 
   private node<T extends Node>(name: string | T): T {
@@ -1011,7 +1134,6 @@ export class HTTP {
     return res;
   }
 
-  // TODO(indutny): use type for `name`
   private invokePausable(name: string, errorCode: ERROR, next: string | Node)
     : Node {
     let cb;
@@ -1026,6 +1148,12 @@ export class HTTP {
       case 'on_status_complete':
         cb = this.callback.onStatusComplete;
         break;
+      case 'on_method_complete':
+        cb = this.callback.onMethodComplete;
+        break;
+      case 'on_version_complete':
+        cb = this.callback.onVersionComplete;
+        break;
       case 'on_header_field_complete':
         cb = this.callback.onHeaderFieldComplete;
         break;
@@ -1038,9 +1166,18 @@ export class HTTP {
       case 'on_chunk_header':
         cb = this.callback.onChunkHeader;
         break;
+      case 'on_chunk_extension_name':
+        cb = this.callback.onChunkExtensionName;
+        break;
+      case 'on_chunk_extension_value':
+        cb = this.callback.onChunkExtensionValue;
+        break;
       case 'on_chunk_complete':
         cb = this.callback.onChunkComplete;
         break;
+      case 'on_reset':
+        cb = this.callback.onReset;
+        break;
       default:
         throw new Error('Unknown callback: ' + name);
     }
